"""Permission storage model."""
import logging
from enum import IntEnum
from functools import reduce
from typing import List, Optional, Union

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, User
from django.db import models, transaction

logger = logging.getLogger(__name__)
PermissionList = List["Permission"]
UserOrGroup = Union[User, Group]


# Get and store the anonymous user for later use to avoid hits to the database.
ANONYMOUS_USER = None


def get_anonymous_user() -> User:
    """Get the anonymous user.

    Note that is the actual user object with username specified in setting
    ANONYMOUS_USER_NAME or id specified in setting ANONYMOUS_USER_ID. The later
    setting has precedence.

    Store the computed value into global variable get_anonymous_user() to avoid
    querying the database every time.

    :raises RuntimeError: when ANONYMOUS_USER_NAME and ANONYMOUS_USER_ID are
        not set.
    """
    global ANONYMOUS_USER
    if ANONYMOUS_USER is None:
        ANONYMOUS_USER = get_user_model().objects.get(
            username=settings.ANONYMOUS_USER_NAME
        )
    return ANONYMOUS_USER


class Permission(IntEnum):
    """Enum that describes all possible permissions on Resolwe objects.

    The permissions on Resolwe objects are ordered linearly so they can be
    mapped to natural numbers. Permissions that are mapped to lower numbers are
    implicitely included in the permissions mapped to a higher number.

    Whenever dealing with permissions (reading, setting...) it is necessary to
    use this enum.

    The instances of Permission class are iterable and iterating over an
    instance P returs all permissions that are lower or equal to P, ordered
    from bottom to top.
    """

    # Possible permission levels. They must be orded from bottom to top since
    # methods bellow rely on the ordering.
    NONE = 0
    VIEW = 1
    EDIT = 2
    SHARE = 4
    OWNER = 8

    @staticmethod
    def from_name(permission_name: str) -> "Permission":
        """Get the permission from permission name.

        :returns: Permission object when permission_name is the name of the
            permission and highest permission if permission_name is 'ALL'.

        :raises KeyError: when permission name is not known.
        """
        if permission_name == "ALL":
            return Permission.highest()
        return Permission[permission_name.upper()]

    @staticmethod
    def highest() -> "Permission":
        """Return the highest permission."""
        return list(Permission)[-1]

    def __iter__(self):
        """Iterate through permission in increasing order.

        When iterating over permission instance the permissions included in the
        current one (including permission itself) are returned (in increasing
        order).

        The permission Permission.NONE is excluded from the listing.
        """
        for permission in Permission:
            if 0 < permission.value <= self.value:
                yield permission

    def __reversed__(self):
        """Iterate through permissions in decreasing order.

        When iterating over permission instance the permissions included in the
        current one (including permission itself) are returned (in decreasing
        order).

        The permission Permission.NONE is excluded from the listing.
        """
        for permission in reversed(Permission):
            if 0 < permission.value <= self.value:
                yield permission

    def __str__(self) -> str:
        """Get the string representation of the permission.

        This is used in serialization so it must be equal to 'view', 'edit',
        'share' and 'owner'.
        """
        return self.name.lower()


class PositivePermissionsManager(models.Manager):
    """Return only PermissionModels with values greater than 0."""

    def get_queryset(self) -> models.QuerySet:
        """Override default queryset."""
        return super().get_queryset().filter(value__gt=0)


class PermissionModel(models.Model):
    """Store a permission for a singe user/group on permission group.

    Exactly one of fields user/group must be non-null.
    """

    # Access all permissions, including the ones with 0 permission.
    all_objects = models.Manager()
    # By default return only positive permissions.
    objects = PositivePermissionsManager()

    #: permission value
    value = models.PositiveSmallIntegerField()

    #: user this permission belongs to
    user = models.ForeignKey(
        get_user_model(),
        related_name="model_permissions",
        on_delete=models.CASCADE,
        null=True,
    )

    #: group this permission belongs to
    group = models.ForeignKey(
        Group, related_name="model_permissions", on_delete=models.CASCADE, null=True
    )

    #: permission group this permission belongs to
    permission_group = models.ForeignKey(
        "PermissionGroup", on_delete=models.CASCADE, related_name="permissions"
    )

    class Meta:
        """Define constraints enforced on the model.

        The tripple (value, permission_group, user/group) must be unique.
        """

        constraints = [
            models.UniqueConstraint(
                fields=["permission_group", "value", "user"],
                name="one_permission_per_user",
                condition=models.Q(user__isnull=False),
            ),
            models.UniqueConstraint(
                fields=["permission_group", "value", "group"],
                name="one_permission_per_group",
                condition=models.Q(group__isnull=False),
            ),
            models.CheckConstraint(
                check=models.Q(user__isnull=False, group__isnull=True)
                | models.Q(user__isnull=True, group__isnull=False),
                name="exactly_one_of_user_group_must_be_set",
            ),
        ]

    @property
    def permission(self) -> Permission:
        """Return the permission object associated with this instance."""
        return Permission(self.value)

    @property
    def permissions(self) -> PermissionList:
        """Return the permission objects associated with this instance."""
        return list(self.permission)

    def __str__(self) -> str:
        """Get the string representation used for debugging."""
        return (
            f"PermissionModel({self.id}, {Permission(self.permission)}, "
            f"user: {self.user}, group: {self.group})"
        )


class PermissionGroup(models.Model):
    """Group of objecs that have the same permissions.

    Example: a container and all its contents have same permission group.
    """

    def __str__(self) -> str:
        """Return the string representation used for debugging purposes."""
        return f"PermissionGroup({self.id}, {self.permissions.all()})"

    @transaction.atomic
    def set_permission(self, permission: Permission, user_or_group: UserOrGroup):
        """Set the given permission on this permission group.

        All previous permissions are removed.

        :raises AssertionError: when no user or group is given.
        """
        from resolwe.permissions.utils import get_identity  # Circular import

        entity, entity_name = get_identity(user_or_group)
        self.permissions.update_or_create(
            **{"permission_group": self, entity_name: entity},
            defaults={"value": permission.value},
        )

    def get_permission(self, user_or_group: UserOrGroup) -> Permission:
        """Get the permission for the given user or group."""
        from resolwe.permissions.utils import get_identity  # Circular import

        entity, entity_name = get_identity(user_or_group)

        # Superuser has all the permissions.
        if entity_name == "user" and entity.is_superuser:
            return Permission.highest()

        permission_model = self.permissions.filter(**{entity_name: entity}).first()
        if permission_model is not None:
            return permission_model.permission
        else:
            return Permission.NONE


class PermissionQuerySet(models.QuerySet):
    """Queryset with methods that simlify filtering by permissions."""

    def _filter_by_permission(
        self,
        user: Optional[User],
        groups: models.QuerySet,
        permission: Permission,
        public: bool = True,
        with_superuser: bool = True,
    ) -> models.QuerySet:
        """Filter queryset by permissions.

        This is a generic method that is called in public methods.

        :attr user: the user which permissions should be considered.

        :attr groups: the groups which permissions should be considered.

        :attr permission: the lowest permission entity must have.

        :attr public: when True consider also public permission.

        :attr with_superuser: when false treat superuser as reguar user.
        """

        # Skip filtering for superuser when with_superuser is set.
        if user is not None and user.is_superuser and with_superuser:
            return self

        # Handle special case of Storage and Relation.
        filters_prefix = ""
        if self.model._meta.label == "flow.Storage":
            filters_prefix = "data__"

        filters = dict()
        if user:
            filters["user"] = models.Q(
                **{
                    f"{filters_prefix}permission_group__permissions__user": user,
                    f"{filters_prefix}permission_group__permissions__value__gte": permission,
                }
            )

        if public:
            filters["public"] = models.Q(
                **{
                    f"{filters_prefix}permission_group__permissions__user": get_anonymous_user(),
                    f"{filters_prefix}permission_group__permissions__value__gte": permission,
                }
            )
        if groups:
            filters["groups"] = models.Q(
                **{
                    f"{filters_prefix}permission_group__permissions__group__in": groups.values_list(
                        "pk", flat=True
                    ),
                    f"{filters_prefix}permission_group__permissions__value__gte": permission,
                }
            )

        # List here is needed otherwise more joins are performed on the query
        # bellow. Some Django queries (for example ExpressionLateralJoin) do
        # not like that and will fail without evaluating the ids query first.
        ids = list(
            self.filter(
                reduce(lambda filters, filter: filters | filter, filters.values())
            )
            .distinct()
            .values_list("id", flat=True)
        )
        return self.filter(id__in=ids)

    def filter_for_user(
        self,
        user: User,
        permission: Permission = Permission.VIEW,
        use_groups: bool = True,
        public: bool = True,
        with_superuser: bool = True,
    ) -> models.QuerySet:
        """Filter objects for user.

        :attr user: the user which permissions should be considered.

        :attr permission: the lowest permission entity must have.

        :attr use_groups: when True consider also permissions of the user groups.

        :attr public: when True consider also public permission.

        :attr with_superuser: when false treat superuser as reguar user.
        """

        from resolwe.permissions.utils import get_user  # Circular import

        user = get_user(user)
        groups = user.groups.all() if use_groups else user.groups.none()

        return self._filter_by_permission(
            user, groups, permission, public, with_superuser
        )


class PermissionObject(models.Model):
    """Base permission object.

    Every object that has permissions must inherit from this one.
    """

    class Meta:
        """Make this class abstract so no new table is created for it."""

        abstract = True

    #: permission group for the object
    permission_group = models.ForeignKey(
        PermissionGroup, on_delete=models.CASCADE, related_name="%(class)s", null=True
    )

    #: custom manager with permission filtering methods.
    objects = PermissionQuerySet.as_manager()

    def __init__(self, *args, **kwargs):
        """Initialize."""
        # The properties used to determine if object is in container.
        self._container_properties = ("collection", "entity")
        super().__init__(*args, **kwargs)

    def is_owner(self, user: User) -> bool:
        """Return if user is the owner of this instance."""
        return self.has_permission(Permission.OWNER, user)

    def has_permission(self, permission: Permission, user: User):
        """Check if user has the given permission on the current object."""
        return (
            self._meta.model.objects.filter(pk=self.pk)
            .filter_for_user(user, permission)
            .exists()
        )

    def set_permission(self, permission: Permission, user_or_group: UserOrGroup):
        """Set permission on this instance.

        It performs additional check if permissions can be set on this object.

        :raises RuntimeError: when given object has no permission_group or is
            contained in a container.
        """
        if self.in_container():
            raise RuntimeError(
                f"The permissions can not be set on object {self} ({self._meta.label}) in container."
            )

        self.permission_group.set_permission(permission, user_or_group)

    def get_permission(self, user_or_group: UserOrGroup) -> Permission:
        """Get permission for given user or group on this instance."""
        return self.permission_group.get_permission(user_or_group)

    def get_permissions(self, user_or_group: UserOrGroup) -> PermissionList:
        """Get a list of all permissions on the object.

        The Permission.NONE is excluded from list.
        """
        permission = self.get_permission(user_or_group)
        if permission:
            return list(permission)
        else:
            return []

    @property
    def topmost_container(self) -> Optional[models.Model]:
        """Get the top-most container of the object.

        :returns: the top-most container or None if it does not exist.
        """
        for property in self._container_properties:
            value = getattr(self, property, None)
            if value is not None:
                return value
        return None

    def in_container(self) -> bool:
        """Return if object lies in a container."""
        return self.topmost_container is not None

    def save(self, *args, **kwargs):
        """Set the permission_group property of the object.

        If object belongs to a container set its permission group to the one
        used by the container.

        If object does not belong to a container and has no permission group
        create a new one and assign it to the object.

        Mental note: this could lead to orphaned PermissionGroup objects lying
        around.
        """
        container = self.topmost_container
        if container is not None:
            self.permission_group_id = container.permission_group_id
        elif self.permission_group is None:
            self.permission_group = PermissionGroup.objects.create()
        # The instance was not yet commited to the database and permission
        # group attribute has been set.
        # Be careful: this can happen on rollback due to slug colision on base
        # object. In this case the property is set to a permission group, which
        # is not commited to the database due to slug colision.
        # We have to create a new one.
        elif self.pk is None:
            try:
                self.permission_group.refresh_from_db()
            except PermissionGroup.DoesNotExist:
                # Instance was not commited to the database.
                self.permission_group = PermissionGroup.objects.create()

        super().save(*args, **kwargs)



"""Resolwe descriptor schema model."""
from django.db import models

from resolwe.permissions.models import PermissionObject

from .base import BaseModel


class DescriptorSchema(BaseModel, PermissionObject):
    """Postgres model for storing descriptors."""

    class Meta(BaseModel.Meta):
        """DescriptorSchema Meta options."""

        permissions = (
            ("view", "Can view descriptor schema"),
            ("edit", "Can edit descriptor schema"),
            ("share", "Can share descriptor schema"),
            ("owner", "Is owner of the description schema"),
        )

    #: detailed description
    description = models.TextField(blank=True)

    #: user descriptor schema represented as a JSON object
    schema = models.JSONField(default=list)



"""Listener plugin and plugin manager class."""
import abc
import logging
from typing import TYPE_CHECKING, Awaitable, Callable, Dict, Optional, Set, Type

from resolwe.flow.executors.socket_utils import Message, Response
from resolwe.utils import BraceMessage as __

if TYPE_CHECKING:
    from resolwe.flow.managers.listener.listener import Processor  # noqa: F401


logger = logging.getLogger(__name__)


class ListenerPlugin(metaclass=abc.ABCMeta):
    """Base class for listener plugin classes."""

    # Prefix used in the names of the methods that handle commands.
    handler_prefix = "handle_"

    @classmethod
    def __init_subclass__(cls: Type["ListenerPlugin"], **kwargs):
        """Register class with the registry on initialization."""
        super().__init_subclass__(**kwargs)
        plugin_manager.add_plugin(cls)

    def _get_handler_name(self, command_name: str) -> str:
        """Get the handler name."""
        return f"{self.handler_prefix}{command_name}"

    def get_handler(
        self, command_name: str
    ) -> Optional[Callable[[Message, "Processor"], Response]]:
        """Return the plugin handler.

        When no such plugin exists return None.
        """
        return getattr(self, self._get_handler_name(command_name), None)

    def get_handled_commands(self) -> Set[str]:
        """Return handled command names."""

        return {
            method_name[len(self.handler_prefix) :]
            for method_name in dir(self)
            if (
                method_name.startswith(self.handler_prefix)
                and callable(getattr(self, method_name))
            )
        }


class ListenerPlugins:
    """Class for registering listener plugins."""

    def __init__(self):
        """Initialize."""
        self._plugin_instances: Dict[str, ListenerPlugin] = dict()
        self._command_handlers: Dict[
            str, Callable[[Dict, "Processor"], Awaitable[Dict]]
        ] = dict()
        self._handler_names = set()

    def _get_class_name(self, plugin_class: Type[ListenerPlugin]) -> str:
        """Get the full class name."""
        return "{}.{}".format(plugin_class.__module__, plugin_class.__name__)

    def add_plugin(self, plugin_class: Type[ListenerPlugin]):
        """Add new plugin class.

        :raises AssertionError: when plugin is already registered.
        """
        plugin_name = self._get_class_name(plugin_class)
        assert (
            plugin_name not in self._plugin_instances
        ), "Plugin named {plugin_name} already registered."
        plugin_instance = plugin_class()
        # Sanity check before we register a plugin.
        for handled_command in plugin_instance.get_handled_commands():
            assert (
                handled_command not in self._command_handlers
            ), "Command {handled_command} is already registered"

        for command_name in plugin_instance.get_handled_commands():
            self._command_handlers[command_name] = plugin_instance.get_handler(
                command_name
            )
        self._plugin_instances[plugin_name] = plugin_instance
        logger.debug(__("Listener plugin {} was registered.", plugin_name))

    def get_handler(self, command_name: str) -> Optional[Callable[[Message], Response]]:
        """Return the handler for the given command.

        When such handler exists in multiple plugins one is chosen at random.
        """
        return self._command_handlers.get(command_name)


plugin_manager = ListenerPlugins()



"""Resolwe data serializer."""
from django.db import transaction

from rest_framework import serializers

from resolwe.flow.models import Collection, Data, DescriptorSchema, Entity, Process
from resolwe.permissions.models import Permission
from resolwe.rest.fields import ProjectableJSONField

from .base import ResolweBaseSerializer
from .collection import CollectionSerializer
from .descriptor import DescriptorSchemaSerializer
from .entity import EntitySerializer
from .fields import DictRelatedField
from .process import ProcessSerializer


class DataSerializer(ResolweBaseSerializer):
    """Serializer for Data objects."""

    input = ProjectableJSONField(required=False)
    output = ProjectableJSONField(required=False)
    descriptor = ProjectableJSONField(required=False)
    process_resources = ProjectableJSONField(required=False)
    process = DictRelatedField(
        queryset=Process.objects.all(), serializer=ProcessSerializer
    )
    descriptor_schema = DictRelatedField(
        queryset=DescriptorSchema.objects.all(),
        serializer=DescriptorSchemaSerializer,
        allow_null=True,
        required=False,
    )
    collection = DictRelatedField(
        queryset=Collection.objects.all(),
        serializer=CollectionSerializer,
        allow_null=True,
        required=False,
        write_permission=Permission.EDIT,
    )
    entity = DictRelatedField(
        queryset=Entity.objects.all(),
        serializer=EntitySerializer,
        allow_null=True,
        required=False,
        write_permission=Permission.EDIT,
    )

    class Meta:
        """DataSerializer Meta options."""

        model = Data
        read_only_fields = (
            "checksum",
            "created",
            "descriptor_dirty",
            "duplicated",
            "finished",
            "id",
            "modified",
            "process_cores",
            "process_error",
            "process_info",
            "process_memory",
            "process_progress",
            "process_rc",
            "process_warning",
            "output",
            "scheduled",
            "size",
            "started",
            "status",
        )
        update_protected_fields = (
            "contributor",
            "input",
            "process",
        )
        fields = (
            read_only_fields
            + update_protected_fields
            + (
                "collection",
                "descriptor",
                "descriptor_schema",
                "entity",
                "name",
                "slug",
                "tags",
                "process_resources",
            )
        )

    def validate_process(self, process):
        """Check that process is active."""
        if not process.is_active:
            raise serializers.ValidationError(
                "Process {} is not active.".format(process)
            )
        return process

    def validate_collection(self, collection):
        """Verify that changing collection is done in the right place."""
        if self.instance and self.instance.collection != collection:
            self.instance.validate_change_collection(collection)
        return collection

    @transaction.atomic
    def update(self, instance, validated_data):
        """Update."""
        update_collection = "collection" in validated_data
        new_collection = validated_data.pop("collection", None)
        instance = super().update(instance, validated_data)
        if update_collection and new_collection != instance.collection:
            instance.move_to_collection(new_collection)
        return instance



"""Support for selective field serialization."""
from rest_framework.fields import JSONField

from .projection import apply_subfield_projection


class ProjectableJSONField(JSONField):
    """JSON field which supports projection."""

    def to_representation(self, value):
        """Project outgoing native value."""
        value = apply_subfield_projection(self, value, deep=True)
        return super().to_representation(value)



"""Application configuration."""
import logging
from contextlib import suppress
from importlib import import_module

from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _

from resolwe.storage.connectors import connectors
from resolwe.storage.connectors.baseconnector import BaseStorageConnector
from resolwe.storage.settings import STORAGE_CONNECTORS

logger = logging.getLogger(__name__)


class StorageConfig(AppConfig):
    """Application configuration."""

    name = "resolwe.storage"
    verbose_name = _("Resolwe Storage")

    def _check_connector_settings(self):
        """Validate the storage connector settings in the django config.

        When there exists a section that does not match any known storage
        connector then error is logged.
        """
        for connector_name, connector_settings in STORAGE_CONNECTORS.items():
            if connector_name not in connectors:
                full_class_name = connector_settings.get("connector")
                class_exists = False
                is_subclass = False
                with suppress(Exception):
                    module_name, class_name = full_class_name.rsplit(".", 1)
                    module = import_module(module_name)
                    class_exists = hasattr(module, class_name)
                    if class_exists:
                        is_subclass = issubclass(
                            getattr(module, class_name), BaseStorageConnector
                        )
                message = "Connector named {} using class {} is not registered.".format(
                    connector_name, full_class_name
                )
                if not class_exists:
                    message += " Class does not exist."
                elif not is_subclass:
                    message += " Class is not a subclass of BaseStorageConnector."
                logger.warning(message)

    def ready(self):
        """Application initialization."""
        self._check_connector_settings()
        # Register signals handlers
        from . import signals  # noqa: F401

        return super().ready()



# pylint: disable=missing-docstring
from pathlib import Path
from unittest.mock import patch

from resolwe.flow.executors import constants
from resolwe.flow.executors.socket_utils import Message, Response, ResponseStatus
from resolwe.flow.managers.listener.basic_commands_plugin import BasicCommands
from resolwe.flow.managers.listener.listener import Processor
from resolwe.flow.managers.protocol import ExecutorProtocol
from resolwe.flow.models import Data, DataDependency, Worker
from resolwe.storage.connectors.baseconnector import BaseStorageConnector
from resolwe.storage.models import FileStorage, ReferencedPath, StorageLocation
from resolwe.test import TestCase


class ListenerTest(TestCase):
    fixtures = ["storage_data.yaml", "storage_processes.yaml", "storage_users.yaml"]

    @classmethod
    def setUpTestData(cls):
        cls.manager = Processor(b"1", 1, None)
        cls.processor = BasicCommands()
        cls.file_storage = FileStorage.objects.get(id=1)
        cls.storage_location = StorageLocation.objects.create(
            file_storage=cls.file_storage, connector_name="GCS", status="OK"
        )
        cls.path = ReferencedPath.objects.create(
            path="test.me", md5="md5", crc32c="crc", awss3etag="aws"
        )
        cls.storage_location.files.add(cls.path)

    def test_handle_download_finished_missing_storage_location(self):
        obj = Message.command(ExecutorProtocol.DOWNLOAD_FINISHED, -2)
        with self.assertRaises(StorageLocation.DoesNotExist):
            self.processor.handle_download_finished(obj, self.manager)

    def test_handle_download_finished(self):
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage, connector_name="local"
        )
        obj = Message.command(ExecutorProtocol.DOWNLOAD_FINISHED, storage_location.id)

        with patch(
            "resolwe.storage.models.FileStorage.default_storage_location",
            self.storage_location,
        ):
            response = self.processor.handle_download_finished(obj, self.manager)

        self.assertEqual(response.response_status, ResponseStatus.OK)
        storage_location.refresh_from_db()
        self.assertEqual(storage_location.status, StorageLocation.STATUS_DONE)
        self.assertEqual(storage_location.files.count(), 1)
        file = storage_location.files.get()
        self.assertEqual(file.path, "test.me")
        self.assertEqual(file.md5, "md5")
        self.assertEqual(file.crc32c, "crc")
        self.assertEqual(file.awss3etag, "aws")

    def test_handle_download_aborted_missing_storage_location(self):
        obj = Message.command(ExecutorProtocol.DOWNLOAD_ABORTED, -2)
        response = self.processor.handle_download_aborted(obj, self.manager)
        self.assertEqual(response.response_status, ResponseStatus.OK)

    def test_handle_download_aborted(self):
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage,
            connector_name="local",
            status=StorageLocation.STATUS_UPLOADING,
        )
        obj = Message.command(ExecutorProtocol.DOWNLOAD_ABORTED, storage_location.id)
        self.processor.handle_download_aborted(obj, self.manager)

        storage_location.refresh_from_db()
        self.assertEqual(storage_location.status, StorageLocation.STATUS_PREPARING)

    def test_handle_download_started_no_location(self):
        obj = Message.command(
            ExecutorProtocol.DOWNLOAD_STARTED,
            {
                "storage_location_id": -2,
                "download_started_lock": True,
            },
        )
        with self.assertRaises(StorageLocation.DoesNotExist):
            self.processor.handle_download_started(obj, self.manager)

    def test_handle_download_started_ok_no_lock_preparing(self):
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage, connector_name="local"
        )

        obj = Message.command(
            ExecutorProtocol.DOWNLOAD_STARTED,
            {
                "storage_location_id": storage_location.id,
                "download_started_lock": False,
            },
        )
        response = self.processor.handle_download_started(obj, self.manager)
        self.assertEqual(
            response, Response(ResponseStatus.OK.value, "download_started")
        )
        storage_location.refresh_from_db()
        self.assertEqual(storage_location.status, StorageLocation.STATUS_PREPARING)

    def test_handle_download_started_ok_no_lock_uploading(self):
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage,
            connector_name="local",
            status=StorageLocation.STATUS_UPLOADING,
        )
        obj = Message.command(
            ExecutorProtocol.DOWNLOAD_STARTED,
            {
                "storage_location_id": storage_location.id,
                "download_started_lock": False,
            },
        )

        response = self.processor.handle_download_started(obj, self.manager)
        self.assertEqual(
            response, Response(ResponseStatus.OK.value, "download_in_progress")
        )
        storage_location.refresh_from_db()
        self.assertEqual(storage_location.status, StorageLocation.STATUS_UPLOADING)

    def test_handle_download_started_ok_no_lock_done(self):
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage,
            connector_name="local",
            status=StorageLocation.STATUS_DONE,
        )
        obj = Message.command(
            ExecutorProtocol.DOWNLOAD_STARTED,
            {
                "storage_location_id": storage_location.id,
                "download_started_lock": False,
            },
        )
        response = self.processor.handle_download_started(obj, self.manager)
        self.assertEqual(
            response, Response(ResponseStatus.OK.value, "download_finished")
        )
        storage_location.refresh_from_db()
        self.assertEqual(storage_location.status, StorageLocation.STATUS_DONE)

    def test_handle_download_started_ok_lock(self):
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage, connector_name="local"
        )

        obj = Message.command(
            ExecutorProtocol.DOWNLOAD_STARTED,
            {
                "storage_location_id": storage_location.id,
                "download_started_lock": True,
            },
        )
        response = self.processor.handle_download_started(obj, self.manager)
        self.assertEqual(
            response, Response(ResponseStatus.OK.value, "download_started")
        )
        storage_location.refresh_from_db()
        self.assertEqual(storage_location.status, StorageLocation.STATUS_UPLOADING)

    def test_handle_get_files_to_download_missing_storage_location(self):
        obj = Message.command(ExecutorProtocol.GET_FILES_TO_DOWNLOAD, -2)
        response = self.processor.handle_get_files_to_download(obj, self.manager)
        self.assertEqual(response, Response(ResponseStatus.OK.value, []))

    def test_handle_get_files_to_download(self):
        obj = Message.command(
            ExecutorProtocol.GET_FILES_TO_DOWNLOAD, self.storage_location.id
        )
        response = self.processor.handle_get_files_to_download(obj, self.manager)
        expected = Response(
            ResponseStatus.OK.value,
            [
                {
                    "id": self.path.id,
                    "path": "test.me",
                    "size": -1,
                    "md5": "md5",
                    "crc32c": "crc",
                    "awss3etag": "aws",
                    "chunk_size": BaseStorageConnector.CHUNK_SIZE,
                }
            ],
        )
        self.assertEqual(response, expected)

    def test_handle_get_referenced_files(self):
        obj = Message.command(ExecutorProtocol.GET_REFERENCED_FILES, "")
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage,
            connector_name="local",
            status=StorageLocation.STATUS_DONE,
            url=str(self.file_storage.id),
        )
        path = Path(storage_location.get_path(filename="output.txt"))
        path.parent.mkdir(exist_ok=True, parents=True)
        path.touch()
        data = Data.objects.get(id=1)
        data.process.output_schema = [{"name": "output_file", "type": "basic:file:"}]
        data.process.save()
        data.output = {"output_file": {"file": "output.txt"}}
        data.save()

        response = self.processor.handle_get_referenced_files(obj, self.manager)
        expected = Response(
            ResponseStatus.OK.value,
            [
                "jsonout.txt",
                "stdout.txt",
                "output.txt",
            ],
        )
        self.assertEqual(response, expected)

    def test_handle_missing_data_locations_missing_data(self):
        obj = Message.command(ExecutorProtocol.MISSING_DATA_LOCATIONS, "")
        response = self.processor.handle_missing_data_locations(obj, self.manager)
        self.assertEqual(response, Response(ResponseStatus.OK.value, {}))

    def test_handle_missing_data_locations_missing_storage_location(self):
        obj = Message.command(ExecutorProtocol.MISSING_DATA_LOCATIONS, "")
        parent = Data.objects.get(id=2)
        child = Data.objects.get(id=1)
        DataDependency.objects.create(
            parent=parent, child=child, kind=DataDependency.KIND_IO
        )
        response = self.processor.handle_missing_data_locations(obj, self.manager)
        expected = Response(ResponseStatus.ERROR.value, "No storage location exists")
        self.assertEqual(response, expected)
        self.assertEqual(StorageLocation.all_objects.count(), 1)

    def test_handle_missing_data_locations_none(self):
        obj = Message.command(ExecutorProtocol.MISSING_DATA_LOCATIONS, "")
        parent = Data.objects.get(id=2)
        child = Data.objects.get(id=1)
        DataDependency.objects.create(
            parent=parent, child=child, kind=DataDependency.KIND_IO
        )
        StorageLocation.objects.create(
            file_storage=parent.location,
            connector_name="local",
            status=StorageLocation.STATUS_DONE,
            url="url",
        )
        response = self.processor.handle_missing_data_locations(obj, self.manager)
        expected = Response(ResponseStatus.OK.value, {})
        self.assertEqual(response, expected)
        self.assertEqual(StorageLocation.all_objects.count(), 2)

    def test_handle_missing_data_locations(self):
        obj = Message.command(ExecutorProtocol.MISSING_DATA_LOCATIONS, "")
        parent = Data.objects.get(id=2)
        child = Data.objects.get(id=1)
        DataDependency.objects.create(
            parent=parent, child=child, kind=DataDependency.KIND_IO
        )
        storage_location = StorageLocation.objects.create(
            file_storage=parent.location,
            connector_name="not_local",
            status=StorageLocation.STATUS_DONE,
            url="url",
        )
        response = self.processor.handle_missing_data_locations(obj, self.manager)
        self.assertEqual(StorageLocation.all_objects.count(), 3)
        created = StorageLocation.all_objects.last()
        expected = Response(
            ResponseStatus.OK.value,
            {
                "url": {
                    "data_id": parent.id,
                    "from_connector": "not_local",
                    "from_storage_location_id": storage_location.id,
                    "to_storage_location_id": created.id,
                    "to_connector": "local",
                }
            },
        )
        self.assertEqual(response, expected)

    def test_handle_resolve_data_path(self):
        """Test data path resolwing."""
        data = Data.objects.get(id=1)
        Worker.objects.get_or_create(data=data, status=Worker.STATUS_PREPARING)
        message = Message.command("resolve_data_path", data.pk)
        response = self.manager.process_command(message)
        assert response.message_data == str(constants.INPUTS_VOLUME)
        connector_name = "local"
        self.storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage, connector_name=connector_name, status="OK"
        )
        response = self.manager.process_command(message)
        self.assertEqual(response.message_data, f"/data_{connector_name}")



""".. Ignore pydocstyle D400.

=======
Resolwe
=======

Open source enterprise dataflow engine in Django.

"""
from resolwe.__about__ import (  # noqa: F401
    __author__,
    __copyright__,
    __email__,
    __license__,
    __summary__,
    __title__,
    __url__,
    __version__,
)



""".. Ignore pydocstyle D400.

=============
Flow Managers
=============

Workflow workload managers.

.. data:: manager

    The global manager instance.

    :type: :class:`~resolwe.flow.managers.dispatcher.Manager`

.. automodule:: resolwe.flow.managers.dispatcher
    :members:
.. automodule:: resolwe.flow.managers.workload_connectors
.. automodule:: resolwe.flow.managers.listener
.. automodule:: resolwe.flow.managers.consumer
    :members:
.. automodule:: resolwe.flow.managers.utils
    :members:

"""
from .dispatcher import Manager
from .listener import ExecutorListener

__all__ = ("manager", "listener")

manager = Manager()
listener = ExecutorListener()



""".. Ignore pydocstyle D400.

=========================
Prepare runtime directory
=========================

"""
import shutil
from importlib.util import find_spec
from pathlib import Path

from django.core.management.base import BaseCommand, CommandError

from resolwe.storage import settings as storage_settings


def copy_dir(source: Path, destination: Path):
    """Copy dir from source to destination.

    The shutil copytree method could be used but it will not overwrite
    directories (option added in Python 3.8, while we support Python 3.6).

    This method iterates through all entries in source directory and copies
    them over to the destination directory (overwriting them if they
    already exist).
    """
    for source_entry in source.rglob("*"):
        relative_source_path = source_entry.relative_to(source)
        destination_entry = destination / relative_source_path
        if source_entry.is_dir():
            destination_entry.mkdir(exist_ok=True, parents=True)
        elif source_entry.is_file() or source_entry.is_symlink():
            destination_entry.parent.mkdir(exist_ok=True, parents=True)
            shutil.copy(source_entry, destination_entry)


def get_source_path(module_name: str) -> Path:
    """Get the source code for the given module.

    :raises RuntimeError: when no source code can be found.
    """
    spec = find_spec(module_name)
    if spec is None or spec.origin is None:
        raise RuntimeError(f"Source code for module {module_name} could not be found.")
    return Path(spec.origin)


class Command(BaseCommand):
    """Prepare runtime directory."""

    help = "Prepare runtime directory."

    def __init__(self, *args, **kwargs):
        """Initialize command."""
        super().__init__(*args, **kwargs)

        if not "runtime" in storage_settings.FLOW_VOLUMES:
            raise CommandError("Volume 'runtime' must be defined in storage settings.")

        self.runtime_dir = Path(
            storage_settings.FLOW_VOLUMES["runtime"]["config"]["path"]
        )

    def add_arguments(self, parser):
        """Command arguments."""
        parser.add_argument(
            "-c",
            "--clear",
            action="store_true",
            dest="clear",
            help="Clear existing files before copying new files.",
        )

        parser.add_argument(
            "--noinput",
            "--no-input",
            action="store_false",
            dest="interactive",
            help="Do NOT prompt the user for input of any kind.",
        )

    def get_confirmation(self):
        """Get user confirmation to proceed."""
        if self.clear:
            action = "This will DELETE ALL FILES in this location!"
        else:
            action = "This will overwrite existing files!"

        message = (
            "\n"
            "You have requested to prepare runtime directory at the destination\n"
            "location as specified in your settings\n"
            "\n"
            "    {destination}\n"
            "\n"
            "{action}\n"
            "Are you sure you want to do this?\n"
            "\n"
            "Type 'yes' to continue, or 'no' to cancel: ".format(
                destination=self.runtime_dir,
                action=action,
            )
        )

        if input("".join(message)) != "yes":
            raise CommandError("Preparing runtime directory cancelled.")

    def clear_runtime(self):
        """Delete contents of the runtime directory."""
        self.stdout.write("Deleting contents of '{}'.".format(self.runtime_dir))
        for entry in self.runtime_dir.glob("*"):
            if entry.is_file() or entry.is_symlink():
                entry.unlink()
            elif entry.is_dir():
                shutil.rmtree(entry)

    def prepare_runtime(self, runtime_dir):
        """Prepare runtime directory."""
        modules_destination = {
            "resolwe.flow.executors": Path("executors"),
            "resolwe.storage.connectors": Path("executors/connectors"),
        }

        for module_name, relative_destination in modules_destination.items():
            copy_dir(
                get_source_path(module_name).parent,
                runtime_dir / relative_destination,
            )

        # Copy files necessary to bootstrap python process to the executor.
        # These files must be mounted inside container, the rest of the runtime
        # is transfered over the protocol.
        for file in [
            "resolwe.process.communicator",
            "resolwe.process.bootstrap_python_runtime",
        ]:
            shutil.copy(get_source_path(file), runtime_dir)

    def set_options(self, **options):
        """Set instance variables based on an options dict."""
        self.clear = options["clear"]
        self.interactive = options["interactive"]

    def handle(self, **options):
        """Collect tools."""
        self.set_options(**options)

        previous_data = any(self.runtime_dir.glob("*"))
        if self.interactive and previous_data:
            self.get_confirmation()

        if self.clear and previous_data:
            self.clear_runtime()

        self.prepare_runtime(self.runtime_dir)



# Generated by Django 3.1.7 on 2021-11-04 11:13

import django.db.models.manager
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("permissions", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="permissionmodel",
            managers=[
                ("all_objects", django.db.models.manager.Manager()),
            ],
        ),
    ]



"""Reslowe process model."""
import copy
import enum
import json
import logging

import jsonschema

from django.contrib.postgres.fields import ArrayField
from django.contrib.postgres.indexes import GinIndex
from django.contrib.postgres.search import SearchVectorField
from django.core.exceptions import PermissionDenied, ValidationError
from django.core.validators import RegexValidator
from django.db import models, transaction

from resolwe.flow.expression_engines.exceptions import EvaluationError
from resolwe.flow.models.utils import (
    DirtyError,
    fill_with_defaults,
    validate_schema,
    validation_schema,
)
from resolwe.flow.utils import dict_dot, get_data_checksum, iterate_fields
from resolwe.permissions.models import PermissionObject, PermissionQuerySet
from resolwe.permissions.utils import assign_contributor_permissions, copy_permissions

from .base import BaseModel, BaseQuerySet
from .descriptor import DescriptorSchema
from .entity import Entity
from .secret import Secret
from .storage import Storage
from .utils import (
    bulk_duplicate,
    hydrate_input_references,
    hydrate_size,
    render_descriptor,
    render_template,
)
from .worker import Worker

# Compatibility for Python < 3.5.
if not hasattr(json, "JSONDecodeError"):
    json.JSONDecodeError = ValueError

logger = logging.getLogger(__name__)


@enum.unique
class HandleEntityOperation(enum.Enum):
    """Constants for entity handling."""

    CREATE = "CREATE"
    ADD = "ADD"
    PASS = "PASS"


class DataQuerySet(BaseQuerySet, PermissionQuerySet):
    """Query set for Data objects."""

    @staticmethod
    def _handle_entity(obj):
        """Create entity if `entity.type` is defined in process.

        Following rules applies for adding `Data` object to `Entity`:
        * Only add `Data object` to `Entity` if process has defined
        `entity.type` field
        * Create new entity if parents do not belong to any `Entity`
        * Add object to existing `Entity`, if all parents that are part
        of it (but not necessary all parents), are part of the same
        `Entity`
        * If parents belong to different `Entities` don't do anything

        """
        entity_type = obj.process.entity_type
        entity_descriptor_schema = obj.process.entity_descriptor_schema
        entity_input = obj.process.entity_input
        entity_always_create = obj.process.entity_always_create
        operation = HandleEntityOperation.PASS

        if entity_type:
            data_filter = {}
            if entity_input:
                input_id = dict_dot(obj.input, entity_input, default=lambda: None)
                if input_id is None:
                    logger.warning("Skipping creation of entity due to missing input.")
                    return
                if isinstance(input_id, int):
                    data_filter["data__pk"] = input_id
                elif isinstance(input_id, list):
                    data_filter["data__pk__in"] = input_id
                else:
                    raise ValueError(
                        "Cannot create entity due to invalid value of field {}.".format(
                            entity_input
                        )
                    )
            else:
                data_filter["data__in"] = obj.parents.all()

            entity_query = Entity.objects.filter(
                type=entity_type, **data_filter
            ).distinct()
            entity_count = entity_query.count()

            if entity_count == 0 or entity_always_create:
                descriptor_schema = DescriptorSchema.objects.filter(
                    slug=entity_descriptor_schema
                ).latest()
                entity = Entity.objects.create(
                    contributor=obj.contributor,
                    descriptor_schema=descriptor_schema,
                    type=entity_type,
                    name=obj.name,
                    tags=obj.tags,
                )
                assign_contributor_permissions(entity)
                operation = HandleEntityOperation.CREATE

            elif entity_count == 1:
                entity = entity_query.first()
                obj.tags = entity.tags
                copy_permissions(entity, obj)
                operation = HandleEntityOperation.ADD

            else:
                logger.info(
                    "Skipping creation of entity due to multiple entities found."
                )
                entity = None

            if entity:
                obj.entity = entity
                obj.save()

            return operation

    @staticmethod
    def _handle_collection(obj, entity_operation=None):
        """Correctly assign Collection to Data and it's Entity.

        There are 2 x 4 possible scenarios how to handle collection
        assignment. One dimension in "decision matrix" is Data.collection:

            1.x Data.collection = None
            2.x Data.collection != None

        Second dimension is about Data.entity:

            x.1 Data.entity is None
            x.2 Data.entity was just created
            x.3 Data.entity already exists and Data.entity.collection = None
            x.4 Data.entity already exists and Data.entity.collection != None
        """
        # 1.2 and 1.3 require no action.

        # 1.1 and 2.1:
        if not obj.entity:
            return
        if entity_operation == HandleEntityOperation.ADD and obj.collection:
            # 2.3
            if not obj.entity.collection:
                raise ValueError(
                    "Created Data has collection {} assigned, but it is added to entity {} that is not "
                    "inside this collection.".format(obj.collection, obj.entity)
                )
            # 2.4
            assert obj.collection == obj.entity.collection

        # 1.4
        if not obj.collection and obj.entity and obj.entity.collection:
            obj.collection = obj.entity.collection
            obj.tags = obj.entity.collection.tags
            obj.save()
        # 2.2
        if entity_operation == HandleEntityOperation.CREATE and obj.collection:
            obj.entity.collection = obj.collection
            obj.entity.tags = obj.collection.tags
            obj.entity.save()
            copy_permissions(obj.collection, obj.entity)

    @transaction.atomic
    def create(self, subprocess_parent=None, **kwargs):
        """Create new object with the given kwargs."""
        obj = super().create(**kwargs)

        # Data dependencies
        obj.save_dependencies(obj.input, obj.process.input_schema)
        if subprocess_parent:
            DataDependency.objects.create(
                parent=subprocess_parent,
                child=obj,
                kind=DataDependency.KIND_SUBPROCESS,
            )
            # Data was from a workflow / spawned process
            if not obj.in_container():
                copy_permissions(subprocess_parent, obj)

        # Entity, Collection assignment
        entity_operation = self._handle_entity(obj)
        self._handle_collection(obj, entity_operation=entity_operation)

        # Assign contributor permission only if Data is not in the container.
        if not obj.in_container():
            assign_contributor_permissions(obj)

        return obj

    @transaction.atomic
    def duplicate(self, contributor, inherit_entity=False, inherit_collection=False):
        """Duplicate (make a copy) ``Data`` objects.

        :param contributor: Duplication user
        """

        return bulk_duplicate(
            data=self,
            contributor=contributor,
            inherit_entity=inherit_entity,
            inherit_collection=inherit_collection,
        )

    @transaction.atomic
    def move_to_collection(self, destination_collection):
        """Move data objects to destination collection.

        Note that this method will also copy tags and permissions
        of the destination collection to the data objects.
        """
        for data in self:
            data.move_to_collection(destination_collection)


class Data(BaseModel, PermissionObject):
    """Postgres model for storing data."""

    class Meta(BaseModel.Meta):
        """Data Meta options."""

        permissions = (
            ("view", "Can view data"),
            ("edit", "Can edit data"),
            ("share", "Can share data"),
            ("owner", "Is owner of the data"),
        )

        indexes = [
            models.Index(name="idx_data_name", fields=["name"]),
            GinIndex(
                name="idx_data_name_trgm", fields=["name"], opclasses=["gin_trgm_ops"]
            ),
            models.Index(name="idx_data_slug", fields=["slug"]),
            models.Index(name="idx_data_status", fields=["status"]),
            GinIndex(name="idx_data_tags", fields=["tags"]),
            GinIndex(name="idx_data_search", fields=["search"]),
        ]

    #: data object is uploading
    STATUS_UPLOADING = "UP"
    #: data object is being resolved
    STATUS_RESOLVING = "RE"
    #: data object is waiting
    STATUS_WAITING = "WT"
    #: data object is preparing
    STATUS_PREPARING = "PP"
    #: data object is processing
    STATUS_PROCESSING = "PR"
    #: data object is done
    STATUS_DONE = "OK"
    #: data object is in error state
    STATUS_ERROR = "ER"
    #: data object is in dirty state
    STATUS_DIRTY = "DR"
    # Assumption (in listener): ordered from least to most problematic.
    STATUS_CHOICES = (
        (STATUS_UPLOADING, "Uploading"),
        (STATUS_RESOLVING, "Resolving"),
        (STATUS_WAITING, "Waiting"),
        (STATUS_PREPARING, "Preparing"),
        (STATUS_PROCESSING, "Processing"),
        (STATUS_DONE, "Done"),
        (STATUS_ERROR, "Error"),
        (STATUS_DIRTY, "Dirty"),
    )

    #: manager
    objects = DataQuerySet.as_manager()

    #: date and time when process was dispatched to the scheduling system
    #: (set by``resolwe.flow.managers.dispatcher.Manager.run``
    scheduled = models.DateTimeField(blank=True, null=True, db_index=True)

    #: process started date and time (set by
    #: ``resolwe.flow.executors.run.BaseFlowExecutor.run`` or its derivatives)
    started = models.DateTimeField(blank=True, null=True, db_index=True)

    #: process finished date date and time (set by
    #: ``resolwe.flow.executors.run.BaseFlowExecutor.run`` or its derivatives)
    finished = models.DateTimeField(blank=True, null=True, db_index=True)

    #: duplication date and time
    duplicated = models.DateTimeField(blank=True, null=True)

    #: checksum field calculated on inputs
    checksum = models.CharField(
        max_length=64,
        db_index=True,
        validators=[
            RegexValidator(
                regex=r"^[0-9a-f]{64}$",
                message="Checksum is exactly 40 alphanumerics",
                code="invalid_checksum",
            )
        ],
    )

    status = models.CharField(
        max_length=2, choices=STATUS_CHOICES, default=STATUS_RESOLVING
    )
    """
    :class:`Data` status

    It can be one of the following:

    - :attr:`STATUS_UPLOADING`
    - :attr:`STATUS_RESOLVING`
    - :attr:`STATUS_WAITING`
    - :attr:`STATUS_PROCESSING`
    - :attr:`STATUS_DONE`
    - :attr:`STATUS_ERROR`
    """

    #: process used to compute the data object
    process = models.ForeignKey("Process", on_delete=models.PROTECT)

    #: process id
    process_pid = models.PositiveIntegerField(blank=True, null=True)

    #: progress
    process_progress = models.PositiveSmallIntegerField(default=0)

    #: return code
    process_rc = models.PositiveSmallIntegerField(blank=True, null=True)

    #: info log message
    process_info = ArrayField(models.CharField(max_length=255), default=list)

    #: warning log message
    process_warning = ArrayField(models.CharField(max_length=255), default=list)

    #: error log message
    process_error = ArrayField(models.CharField(max_length=255), default=list)

    #: actual inputs used by the process
    input = models.JSONField(default=dict)

    #: actual outputs of the process
    output = models.JSONField(default=dict)

    #: total size of data's outputs in bytes
    size = models.BigIntegerField()

    #: data descriptor schema
    descriptor_schema = models.ForeignKey(
        "DescriptorSchema", blank=True, null=True, on_delete=models.PROTECT
    )

    #: actual descriptor
    descriptor = models.JSONField(default=dict)

    #: indicate whether `descriptor` doesn't match `descriptor_schema` (is dirty)
    descriptor_dirty = models.BooleanField(default=False)

    #: track if user set the data name explicitly
    named_by_user = models.BooleanField(default=False)

    #: dependencies between data objects
    parents = models.ManyToManyField(
        "self", through="DataDependency", symmetrical=False, related_name="children"
    )

    #: tags for categorizing objects
    tags = ArrayField(models.CharField(max_length=255), default=list)

    #: actual allocated memory
    process_memory = models.PositiveIntegerField(default=0)

    #: actual allocated cores
    process_cores = models.PositiveSmallIntegerField(default=0)

    #: data location
    location = models.ForeignKey(
        "storage.FileStorage",
        blank=True,
        null=True,
        on_delete=models.PROTECT,
        related_name="data",
    )

    #: entity
    entity = models.ForeignKey(
        "Entity", blank=True, null=True, on_delete=models.CASCADE, related_name="data"
    )

    #: collection
    collection = models.ForeignKey(
        "Collection",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="data",
    )

    #: field used for full-text search
    search = SearchVectorField(null=True)

    #: process requirements overrides
    process_resources = models.JSONField(default=dict)

    def __init__(self, *args, **kwargs):
        """Initialize attributes."""
        super().__init__(*args, **kwargs)
        self._original_name = self.name
        self._original_output = self.output

    def resolve_secrets(self):
        """Retrieve handles for all basic:secret: fields on input.

        The process must have the ``secrets`` resource requirement
        specified in order to access any secrets. Otherwise this method
        will raise a ``PermissionDenied`` exception.

        :return: A dictionary of secrets where key is the secret handle
            and value is the secret value.
        """
        secrets = {}
        for field_schema, fields in iterate_fields(
            self.input, self.process.input_schema
        ):
            if not field_schema.get("type", "").startswith("basic:secret:"):
                continue

            name = field_schema["name"]
            value = fields[name]
            try:
                handle = value["handle"]
            except KeyError:
                continue

            try:
                secrets[handle] = Secret.objects.get_secret(
                    handle, contributor=self.contributor
                )
            except Secret.DoesNotExist:
                raise PermissionDenied(
                    "Access to secret not allowed or secret does not exist"
                )

        # If the process does not not have the right requirements it is not
        # allowed to access any secrets.
        allowed = self.process.requirements.get("resources", {}).get("secrets", False)
        if secrets and not allowed:
            raise PermissionDenied(
                "Process '{}' has secret inputs, but no permission to see secrets".format(
                    self.process.slug
                )
            )

        return secrets

    def save_dependencies(self, instance, schema):
        """Save data: and list:data: references as parents."""

        def add_dependency(value):
            """Add parent Data dependency."""
            try:
                DataDependency.objects.update_or_create(
                    parent=Data.objects.get(pk=value),
                    child=self,
                    defaults={"kind": DataDependency.KIND_IO},
                )
            except Data.DoesNotExist:
                pass

        for field_schema, fields in iterate_fields(instance, schema):
            name = field_schema["name"]
            value = fields[name]

            if field_schema.get("type", "").startswith("data:"):
                add_dependency(value)
            elif field_schema.get("type", "").startswith("list:data:"):
                for data in value:
                    add_dependency(data)

    def save(self, render_name=False, *args, **kwargs):
        """Save the data model."""
        if self.name != self._original_name:
            self.named_by_user = True

        try:
            jsonschema.validate(
                self.process_resources, validation_schema("process_resources")
            )
        except jsonschema.exceptions.ValidationError as exception:
            # Re-raise as Django ValidationError
            raise ValidationError(exception.message)

        create = self.pk is None
        if create:
            fill_with_defaults(self.input, self.process.input_schema)

            if not self.name:
                self._render_name()
            else:
                self.named_by_user = True

            self.checksum = get_data_checksum(
                self.input, self.process.slug, self.process.version
            )

            validate_schema(self.input, self.process.input_schema)

            hydrate_size(self)
            # If only specified fields are updated (e.g. in executor), size needs to be added
            if "update_fields" in kwargs:
                kwargs["update_fields"].append("size")

        elif render_name:
            self._render_name()

        render_descriptor(self)

        if self.descriptor_schema:
            try:
                validate_schema(self.descriptor, self.descriptor_schema.schema)
                self.descriptor_dirty = False
            except DirtyError:
                self.descriptor_dirty = True
        elif self.descriptor and self.descriptor != {}:
            raise ValueError(
                "`descriptor_schema` must be defined if `descriptor` is given"
            )

        with transaction.atomic():
            self._perform_save(*args, **kwargs)

        self._original_output = self.output

    def _perform_save(self, *args, **kwargs):
        """Save the data model."""
        super().save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        """Delete the data model."""
        if hasattr(self, "worker"):
            if self.worker.status not in Worker.FINAL_STATUSES:
                self.worker.terminate()
        # Store ids in memory as relations are also deleted with the Data object.
        storage_ids = list(self.storages.values_list("pk", flat=True))
        super().delete(*args, **kwargs)

        Storage.objects.filter(pk__in=storage_ids, data=None).delete()

    def is_duplicate(self):
        """Return True if data object is a duplicate."""
        return bool(self.duplicated)

    def duplicate(self, contributor, inherit_entity=False, inherit_collection=False):
        """Duplicate (make a copy)."""
        return bulk_duplicate(
            data=self._meta.model.objects.filter(pk=self.pk),
            contributor=contributor,
            inherit_entity=inherit_entity,
            inherit_collection=inherit_collection,
        )[0]

    def move_to_collection(self, collection):
        """Move data object to collection."""
        self.validate_change_collection(collection)
        self.collection = collection
        self.permission_group = collection.permission_group
        if collection:
            self.tags = collection.tags
        self.save(update_fields=["tags", "permission_group", "collection"])

    def move_to_entity(self, entity):
        """Move data object to entity."""
        if entity is None and self.in_container():
            raise ValidationError("Data object must not be removed from container.")
        self.entity = entity
        if entity:
            self.permission_group = entity.permission_group
        self.tags = entity.tags
        self.save(update_fields=["permission_group", "entity", "tags"])

    def validate_change_collection(self, collection):
        """Raise validation error if data object cannot change collection."""
        if self.entity:
            raise ValidationError(
                "If Data is in entity, you can only move it to another collection "
                "by moving entire entity."
            )
        if collection is None:
            raise ValidationError("Data object can not be removed from the container.")

    def _render_name(self):
        """Render data name.

        The rendering is based on name template (`process.data_name`) and
        input context.

        """
        if not self.process.data_name or self.named_by_user:
            return

        inputs = copy.deepcopy(self.input)
        hydrate_input_references(
            inputs, self.process.input_schema, hydrate_values=False
        )
        template_context = inputs

        try:
            name = render_template(
                self.process, self.process.data_name, template_context
            )
        except EvaluationError:
            name = "?"

        self.name = name

    def get_resource_limits(self):
        """Return the resource limits for this data."""
        return self.process.get_resource_limits(self)


class DataDependency(models.Model):
    """Dependency relation between data objects."""

    #: child uses parent's output as its input
    KIND_IO = "io"
    #: child was spawned by the parent
    KIND_SUBPROCESS = "subprocess"
    KIND_DUPLICATE = "duplicate"
    KIND_CHOICES = (
        (KIND_IO, "Input/output dependency"),
        (KIND_SUBPROCESS, "Subprocess"),
        (KIND_DUPLICATE, "Duplicate"),
    )

    #: child data object
    child = models.ForeignKey(
        Data, on_delete=models.CASCADE, related_name="parents_dependency"
    )
    #: parent data object
    parent = models.ForeignKey(
        Data, null=True, on_delete=models.SET_NULL, related_name="children_dependency"
    )
    #: kind of dependency
    kind = models.CharField(max_length=16, choices=KIND_CHOICES)



# Generated by Django 3.1.7 on 2021-11-04 11:13

from django.conf import settings
from django.db import migrations


def create_anonymous_user(apps, schema_editor):
    """Create anonymous user if it does not exist."""
    apps.get_model("auth", "User").objects.get_or_create(
        username=settings.ANONYMOUS_USER_NAME
    )


class Migration(migrations.Migration):

    dependencies = [
        ("permissions", "0003_change_default_manager_on_permission_model"),
    ]

    operations = [
        migrations.RunPython(
            create_anonymous_user, reverse_code=migrations.RunPython.noop
        ),
    ]



""".. Ignore pydocstyle D400.

.. autoclass:: resolwe.flow.executors.local.run.FlowExecutor

"""
import asyncio
import contextlib
import logging
import os
import shlex
from asyncio import subprocess

from ..run import BaseFlowExecutor

logger = logging.getLogger(__name__)


class FlowExecutor(BaseFlowExecutor):
    """Local dataflow executor proxy."""

    name = "local"

    def __init__(self, *args, **kwargs):
        """Initialize attributes."""
        super().__init__(*args, **kwargs)

        self.kill_delay = 5
        self.proc = None
        self.stdout = None
        self.command = "/bin/sh"

    async def start(self):
        """Start process execution."""
        # Workaround for pylint issue #1469
        # (https://github.com/PyCQA/pylint/issues/1469).
        self.proc = await subprocess.create_subprocess_exec(
            *shlex.split(self.command),
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        self.stdout = self.proc.stdout

        return self.proc.pid

    async def run_script(self, script):
        """Execute the script and save results."""
        script = os.linesep.join(["set -x", "set +B", script, "exit"]) + os.linesep
        self.proc.stdin.write(script.encode("utf-8"))
        await self.proc.stdin.drain()
        self.proc.stdin.close()

    async def end(self):
        """End process execution."""
        await self.proc.wait()

        return self.proc.returncode

    async def terminate(self):
        """Terminate a running script."""
        try:
            self.proc.terminate()
        except ProcessLookupError:
            # Process has already been terminated.
            pass
        else:
            with contextlib.suppress(asyncio.TimeoutError):
                await asyncio.wait_for(self.proc.wait(), self.kill_delay)
            if self.proc.returncode is None:
                self.proc.kill()
            await self.proc.wait()

        await super().terminate()



"""Resolwe storage model."""
import copy
import logging
import os
from datetime import datetime, timedelta
from pathlib import PurePath
from typing import Iterable, List, Optional, Union

from django.db import models
from django.utils.timezone import now

from resolwe.flow.models import Data
from resolwe.storage.connectors import DEFAULT_CONNECTOR_PRIORITY, connectors
from resolwe.storage.connectors.baseconnector import BaseStorageConnector
from resolwe.storage.connectors.transfer import Transfer
from resolwe.storage.connectors.utils import paralelize
from resolwe.storage.settings import STORAGE_CONNECTORS

logger = logging.getLogger(__name__)


class FileStorage(models.Model):
    """Proxy between Data object and StorageLocation objects."""

    #: creation date and time
    created = models.DateTimeField(auto_now_add=True)

    def has_storage_location(self, connector_name: str) -> bool:
        """Return if StorageLocation object with given connector name exist.

        Only return StorageLocation objects with status set to OK.
        """
        return self.storage_locations.filter(
            connector_name=connector_name, status=StorageLocation.STATUS_DONE
        ).exists()

    @property
    def default_storage_location(self) -> Optional["StorageLocation"]:
        """Return default storage location for this instance.

        This is a storage location using connector with the lowest priority.
        When multiple storage locations have same priority one of them is
        returned.

        Usually only StorageLocation objects with status OK are considered.
        When no such objects exist one of remaining objects is returned (this
        should only happen when data object is still processing). When no
        StorageLocation objects exist None is returned.

        When StorageLocation has ceonnector that is not defined in settings the
        default priority of 100 is applied to it.

        :returns: instance of StorageLocation class or None.
        :rtype: Optional[StorageLocation]
        """
        whens = [
            models.When(
                connector_name=connector_name,
                then=connectors[connector_name].priority,
            )
            for connector_name in connectors
        ]
        q_set_all = self.storage_locations.annotate(
            priority=models.Case(
                *whens,
                default=DEFAULT_CONNECTOR_PRIORITY,
                output_field=models.IntegerField(),
            )
        ).order_by("priority")
        q_set_done = q_set_all.filter(status=StorageLocation.STATUS_DONE)
        return q_set_done.first() or q_set_all.first()

    @property
    def connectors(self) -> Iterable[BaseStorageConnector]:
        """Get a list of connectors for this location."""
        return [
            connectors[connector_name]
            for connector_name in self.storage_locations.filter(
                status=StorageLocation.STATUS_DONE
            ).values_list("connector_name", flat=True)
        ]

    def get_path(self, prefix: str = None, filename: str = None) -> str:
        """Return path of the default storage location."""
        return self.default_storage_location.get_path(prefix, filename)

    @property
    def subpath(self) -> str:
        """Return URL of the default storage location."""
        return self.default_storage_location.url

    @property
    def files(self) -> models.QuerySet:
        """Get referenced path objects in default storage location."""
        return self.default_storage_location.files

    @property
    def urls(self) -> List[str]:
        """Get a list of URLs of objects stored in default storage location."""
        return self.default_storage_location.urls


class LocationsDoneManager(models.Manager):
    """Return only StorageLocation objects with status set to OK."""

    def get_queryset(self) -> models.QuerySet:
        """Override default queryset."""
        return super().get_queryset().filter(status="OK")

    def _preprocess_delay(self, delay: Optional[int]) -> int:
        """Preprocess delay according to rules.

        1. When delay is None return 0. Entries without explicit delay are
            processed immediately.
        2. When delay is negative return the number of days between the
            current date and 1900. There should be no object created before
            1900 so this means no object will be processed.
        3. Otherwise return the delay.
        """
        if delay is None:
            return 0
        assert isinstance(delay, int), "Delay should be integer or None"
        if delay < 0:
            return (datetime.now() - datetime(1900, 1, 1)).days
        else:
            return delay

    def to_delete(self, connector_name: str) -> models.QuerySet:
        """Get a queryset of locations that must be deleted.

        :returns: a queryset of FileStorage objects whose StorageLocations objects
            belonging to the given connector must be deleted.
        """

        location_settings = copy.deepcopy(STORAGE_CONNECTORS.get(connector_name, {}))
        rules: dict = location_settings.get("config", {}).get("delete", {})

        if not rules:
            return FileStorage.objects.none()

        process_overrides = rules.pop("process_type", {})
        data_slug_overrides = rules.pop("data_slug", {})

        whens = [
            models.When(
                data__slug=data_slug,
                then=timedelta(days=self._preprocess_delay(override_rule.get("delay"))),
            )
            for data_slug, override_rule in data_slug_overrides.items()
        ]
        whens += [
            models.When(
                data__process__type__startswith=process_type,
                then=timedelta(days=self._preprocess_delay(override_rule.get("delay"))),
            )
            for process_type, override_rule in process_overrides.items()
        ]

        connectors_priority_whens = [
            models.When(
                connector_name=connector_name,
                then=connectors[connector_name].priority,
            )
            for connector_name in connectors
        ]

        min_other_copies = rules.get("min_other_copies", 1)
        default_delay = self._preprocess_delay(rules.get("delay"))

        return (
            FileStorage.objects.filter(
                data__status=Data.STATUS_DONE,
                created__lt=now()  # type: ignore
                - models.Case(
                    *whens,
                    default=timedelta(days=default_delay),
                    output_field=models.DurationField(),
                ),
            )
            .annotate(
                all_locations_count=models.Count(
                    "storage_locations",
                    filter=models.Q(
                        storage_locations__status=StorageLocation.STATUS_DONE
                    ),
                    distinct=True,
                )
            )
            .annotate(
                open_logs=models.Exists(
                    AccessLog.objects.filter(
                        storage_location__file_storage=models.OuterRef("id"),
                        storage_location__connector_name=connector_name,
                        finished__isnull=True,
                    )
                ),
                accessed_lately=models.Exists(
                    FileStorage.objects.filter(
                        pk=models.OuterRef("id"),
                        storage_locations__connector_name=connector_name,
                        storage_locations__last_update__gt=now()  # type: ignore
                        - models.Case(
                            *whens,
                            default=timedelta(days=default_delay),
                            output_field=models.DurationField(),
                        ),
                    )
                ),
                default_connector=models.functions.Coalesce(
                    models.Subquery(
                        StorageLocation.objects.filter(
                            file_storage=models.OuterRef("id"),
                            status=StorageLocation.STATUS_DONE,
                        )
                        .annotate(
                            priority=models.Case(
                                *connectors_priority_whens,
                                default=DEFAULT_CONNECTOR_PRIORITY,
                                output_field=models.IntegerField(),
                            )
                        )
                        .order_by("priority")
                        .values_list("connector_name", flat=True)[:1]
                    ),
                    models.Value(""),
                ),
            )
            .filter(
                all_locations_count__gt=min_other_copies,
                open_logs=False,
                accessed_lately=False,
                storage_locations__status=StorageLocation.STATUS_DONE,
                storage_locations__connector_name=connector_name,
            )
            .exclude(default_connector=connector_name)
            .distinct()
        )

    def to_copy(self, connector_name: str) -> models.QuerySet:
        """Get a queryset of locations that must be copied.

        :returns: a queryset of FileStorage objects for which a new copy of
            StorageLocation object belonging to the given connector must be
            created.
        """
        location_settings = copy.deepcopy(STORAGE_CONNECTORS.get(connector_name, {}))
        rules: dict = location_settings.get("config", {}).get("copy", {})

        if not rules:
            return FileStorage.objects.none()

        process_overrides = rules.pop("process_type", {})
        data_slug_overrides = rules.pop("data_slug", {})

        whens = [
            models.When(
                data__slug=data_slug,
                then=timedelta(days=self._preprocess_delay(override_rule.get("delay"))),
            )
            for data_slug, override_rule in data_slug_overrides.items()
        ]
        whens += [
            models.When(
                data__process__type__startswith=process_type,
                then=timedelta(days=self._preprocess_delay(override_rule.get("delay"))),
            )
            for process_type, override_rule in process_overrides.items()
        ]

        default_delay = self._preprocess_delay(rules.get("delay"))

        return (
            FileStorage.objects.filter(data__status=Data.STATUS_DONE)
            .filter(
                created__lt=now()  # type: ignore
                - models.Case(
                    *whens,
                    default=timedelta(days=default_delay),
                    output_field=models.DurationField(),
                )
            )
            .annotate(
                any_location_exists=models.Exists(
                    StorageLocation.objects.filter(file_storage=models.OuterRef("id"))
                ),
                to_location_exists=models.Exists(
                    StorageLocation.objects.filter(
                        file_storage=models.OuterRef("id"),
                        connector_name=connector_name,
                    )
                ),
            )
            .filter(any_location_exists=True, to_location_exists=False)
            .distinct()
        )


class AllLocationsManager(models.Manager):
    """Return all StorageLocation objects."""

    def unreferenced_locations(self) -> models.QuerySet:
        """Get a queryset containing all unreferenced locations.

        The location must also be unlocked (has no accesslogs with finished
        date set to null) to be included in the list.
        """
        return (
            StorageLocation.all_objects.filter(file_storage__data__isnull=True)
            .annotate(
                logs_num=models.Count(
                    "access_logs", filter=models.Q(access_logs__finished__isnull=True)
                )
            )
            .filter(logs_num=0)
        )


class StorageLocation(models.Model):
    """Stores path to where the actual data is stored."""

    # Access all objects through all_objects storage manager.
    # This is also _default_object_manager since it is defined first.
    all_objects = AllLocationsManager()
    # By default iterate only through locations with status set to OK.
    objects = LocationsDoneManager()

    #: data upload is preparing
    STATUS_PREPARING = "PR"
    #: data is uploading
    STATUS_UPLOADING = "UP"
    #: data object is ready
    STATUS_DONE = "OK"
    #: data object is deleting
    STATUS_DELETING = "DE"
    STATUS_CHOICES = (
        (STATUS_PREPARING, "Preparing"),
        (STATUS_UPLOADING, "Uploading"),
        (STATUS_DONE, "Done"),
        (STATUS_DELETING, "Deleting"),
    )

    #: file storage object
    file_storage = models.ForeignKey(
        "FileStorage", on_delete=models.PROTECT, related_name="storage_locations"
    )

    #: url to where data is stored
    url = models.CharField(max_length=60, null=False, blank=False)

    #: name of the connector that manages this location
    connector_name = models.CharField(max_length=30, null=False, blank=False)

    #: date and time of the last update
    last_update = models.DateTimeField(auto_now=True)

    status = models.CharField(
        max_length=2, choices=STATUS_CHOICES, default=STATUS_PREPARING
    )
    """
    :class:`StorageLocation` status

    It can be one of the following:

    - :attr:`STATUS_UPLOADING`
    - :attr:`STATUS_PREPARING`
    - :attr:`STATUS_DONE`
    - :attr:`STATUS_DELETING`
    """

    class Meta:
        """Add unique constaint."""

        unique_together = ("url", "connector_name")

    def __str__(self) -> str:
        """Stringify StorageLocation object."""
        return "StorageLocation(pk={}, connector={}, url={})".format(
            self.pk, self.connector_name, self.url
        )

    def get_path(
        self,
        prefix: Optional[Union[str, PurePath]] = None,
        filename: Optional[Union[str, PurePath]] = None,
    ) -> str:
        """Get the path for this storage location."""
        prefix = prefix or self.connector.base_path
        path = PurePath(prefix) / self.url

        if filename:
            path = path / filename

        return os.fspath(path)

    @property
    def connector(self) -> "BaseStorageConnector":
        """Get the connector for this storage location."""
        return connectors[self.connector_name]

    @property
    def data(self):
        """Return the FileStorage data object."""
        return self.file_storage.data

    @property
    def subpath(self) -> str:
        """Return URL of the default storage location object.

        Exists for backwards compatibility.
        """
        return self.url

    @property
    def locked(self) -> bool:
        """Is this storage location locked."""
        return self.access_logs.filter(finished__isnull=True).exists()

    @property
    def urls(self) -> List[str]:
        """Get a list of URLs of stored files and directories."""
        return [file_.path for file_ in self.files.all()]

    def delete_data(self):
        """Delete all data for this storage location."""
        self.connector.delete(self.url, self.urls)

    def delete(self, *args, **kwargs):
        """Delete StorageLocation object.

        Set status to DELETING, remove data and finally location object.
        """
        # Store a queryset contaning ReferencedPath objects associated with
        # this location. These paths are candidates for deleting.
        referenced_paths = self.files.all()
        self.status = StorageLocation.STATUS_DELETING
        self.save()
        self.delete_data()
        super().delete(*args, **kwargs)
        # Remove ReferencedPath in the stored QuerySet only if they are not
        # referenced by anoter StorageLocation.
        referenced_paths.filter(storage_locations__isnull=True).delete()

    def transfer_data(self, destination: "StorageLocation"):
        """Transfer data to another storage location."""
        t = Transfer(self.connector, destination.connector)
        transfered_files = t.transfer_objects(self.url, list(self.files.values()))

        # When a dictionary in transfered_files list contains the key 'id' it
        # is based on already existing path. In such case we do not have to
        # create a new ReferencedPath object but only add it to the relation.
        new_paths = []
        existing_paths = []
        for file in transfered_files:
            if "id" in file:
                existing_paths.append(file["id"])
            else:
                new_paths.append(ReferencedPath(**file))
        if new_paths:
            ReferencedPath.objects.bulk_create(new_paths)
        destination.files.add(*existing_paths, *new_paths)

    def verify_data(self) -> bool:
        """Verify data stored in this location.

        Verify hashes and sizes of all files. This operation could be slow
        and cause network traffic, use with care.

        :returns: True if data is OK, False otherwise. In case of failure
        additional information about error is logged.
        """

        def worker(referenced_paths: Iterable[ReferencedPath]) -> bool:
            """Check given referenced paths."""
            hash_types = ["md5", "crc32c"]
            connector = self.connector.duplicate()
            for referenced_path in referenced_paths:
                url = self.get_path(filename=referenced_path.path, prefix=PurePath(""))

                # Connector check failed: log error and return False.
                if not connector.check_url(url):
                    logger.error(
                        "Connector {} check for URL {} failed".format(
                            connector.name, url
                        )
                    )
                    return False

                connector_hashes = connector.get_hashes(url, hash_types)
                # Could not retrieve hashes: log error and return False.
                if connector_hashes is None:
                    logger.error(
                        "Connector {} could not retrieve hashes for {}".format(
                            connector.name, url
                        )
                    )
                    return False

                for hash_name, hash_value in connector_hashes.items():
                    referenced_path_hash = getattr(referenced_path, hash_name)
                    if not referenced_path_hash:
                        logger.warning(
                            "ReferencedPath with id {} has no {} hash".format(
                                referenced_path.id, hash_name
                            )
                        )
                        continue

                    # Hashes differ: log error and return False.
                    if referenced_path_hash != hash_value:
                        logger.error(
                            "ReferencedPath with id {} has wrong {} hash: {} instead of {}".format(
                                referenced_path.id,
                                hash_name,
                                referenced_path_hash,
                                hash_value,
                            )
                        )
                        return False
            return True

        max_threads = max(1, min((self.files.count() // 5), 20))
        futures = paralelize(
            self.files.all().exclude(path__endswith="/"),
            worker,
            max_threads=max_threads,
        )
        return all(future.result() for future in futures)


class AccessLog(models.Model):
    """Class that holds access log to storage locations."""

    #: when accessing storage location data started
    started = models.DateTimeField(auto_now_add=True)

    #: when accessing storage location data finished
    finished = models.DateTimeField(null=True, blank=True)

    #: storage location object
    storage_location = models.ForeignKey(
        "StorageLocation", on_delete=models.CASCADE, related_name="access_logs"
    )

    #: human readable reason for access
    reason = models.CharField(max_length=120, null=False, blank=False)

    #: data object that was the cause for the lock
    cause = models.ForeignKey(
        "flow.Data", on_delete=models.CASCADE, related_name="access_logs", null=True
    )


class ReferencedPath(models.Model):
    """Stores reference to a single object (file or directory).

    All references are relative according to the subpath stored in the
    FileStorage object. Directories are stored with '/' at the end and are not
    supported on all conectors.
    """

    #: refers to a file or directory using '/' as separator
    path = models.TextField(db_index=True)

    #: size of the file (-1 undefined)
    size = models.BigIntegerField(default=-1)

    #: StorageLocation objects
    storage_locations = models.ManyToManyField("StorageLocation", related_name="files")

    #: MD5 hexadecimal hash
    md5 = models.CharField(max_length=32, null=False, blank=False)

    #: CRC32C hexadecimal hash
    crc32c = models.CharField(max_length=8, null=False, blank=False)

    #: Etag hexadecimal hash
    awss3etag = models.CharField(max_length=50, null=False, blank=False)

    #: Chunk size used for Etag computation.
    chunk_size = models.IntegerField(default=BaseStorageConnector.CHUNK_SIZE)

    @property
    def file_storage(self) -> Optional[FileStorage]:
        """Get the file storage object that holds this object.

        When no such FileStorage exists None is returned.
        """
        storage_location = self.storage_locations.first()
        if storage_location is None:
            return None

        return storage_location.file_storage
